/**
 * Managing Large Data, Systems Group, ETH Zurich
 * Exercise 7
 * Bitmap Indexing, Query Processing and WAH Compression
 *
 * File        : PlainBitVector.h
 * Author      : Cagri Balkesen, ETH Zurich, 2012-14-05
 *
 */

#ifndef _PLAINBITVECTOR_H_
#define _PLAINBITVECTOR_H_

#include <stdint.h>
#include <vector>
#include <bitset>

class PlainBitVector;

/***** Global bit operations on PlainBitVectors *****/
PlainBitVector operator& (const PlainBitVector& lhs, const PlainBitVector& rhs);
PlainBitVector operator| (const PlainBitVector& lhs, const PlainBitVector& rhs);


/***** PlainBitVector class represents uncompressed bit vectors *****/
class PlainBitVector {
  private:
    uint32_t mSize;
    uint32_t mNumWords;

    /** storage for bits */
    std::vector< std::bitset<32> > mBitWords;

    /** Iterator state variables */
    uint32_t mCurrWordIdx;
    uint32_t mCurrBitIdx;
    uint32_t mCurr;

  public:
    PlainBitVector(uint32_t size) : mSize(size)
      {
          mNumWords = size / 32 + 1;
          /* reserve enough space to avoid re-sizing */
          mBitWords.reserve(mNumWords);
      }

    /** number of bits */
    uint32_t size() const {
        return mSize;
    }

    /** number of 32-bit words required to store the bit vector */
    uint32_t getNumWords() const {
        return this->mNumWords;
    }

    friend PlainBitVector operator& (const PlainBitVector& lhs, const PlainBitVector& rhs);
    friend PlainBitVector operator| (const PlainBitVector& lhs, const PlainBitVector& rhs);

    /**
     * \defgroup BitOperations Bit operation methods over the bit vector.
     * @{ 
     */
    /* In-place AND operator. Assumption vector lengths will be same. */
    const PlainBitVector & operator&= (const PlainBitVector& rhs) {

        for(uint32_t i = 0; i < this->mNumWords; i++) {
            this->mBitWords[i] &= rhs.mBitWords[i];
        }

        return *this;
    }

    /* In-place OR operator. Assumption vector lengths will be same. */
    const PlainBitVector & operator|= (const PlainBitVector& rhs) {

        for(uint32_t i = 0; i < this->mNumWords; i++) {
            this->mBitWords[i] |= rhs.mBitWords[i];
        }

        return *this;

    }

    /** Count set bits in a bit vector */
    uint32_t count() const {

        uint32_t count = 0;

        for(uint32_t i = 0; i < this->mNumWords; i++) {
            count += this->mBitWords[i].count();
        }

        return count;

    }
    /**@}*/

    /** 
     * When building the bitvector, this method is called to inform about the
     * next set-bit's (bit with value 1) index.
     *
     * Subsequent calls of this method must have a greater idx than the previous
     * call. 
     * 
     * @param idx Index of the next set-bit in the bitvector.
     */
    void nextSetBitIndex(uint32_t idx) {
        uint32_t wordIdx = idx / 32;
        uint32_t bitIdx  = idx % 32;

        this->mBitWords[wordIdx].set(bitIdx);
    }

    /**
     * \defgroup BitIteration Iteration methods over the bits of the bit vector.
     * @{ 
     */
    void begin() {
        mCurrWordIdx = 0;
        mCurrBitIdx = 0;
        mCurr = 0;
    }

    /** has more bits? */
    bool hasNext() {
        return (mCurr < mSize);
    }

    /** return next bit, 0:false, 1:true */
    bool getNext() {
        bool res = this->mBitWords[mCurrWordIdx].test(mCurrBitIdx);

        if(mCurrBitIdx == 31) {
            mCurrBitIdx = 0;
            mCurrWordIdx ++;
        }
        else {
            mCurrBitIdx ++;
        }

        mCurr++;

        return res;
    }
    /**@}*/
};

#endif /* _PLAINBITVECTOR_H_ */
