/**
 * Managing Large Data, Systems Group, ETH Zurich
 * Exercise 7
 * Bitmap Indexing, Query Processing and WAH Compression
 *
 * File        : WAHBitVector.h
 * Author      : Cagri Balkesen, ETH Zurich, 2012-14-05
 *
 */

#ifndef _WAHBITVECTOR_H_
#define _WAHBITVECTOR_H_

#include <stdint.h>
#include <vector>
#include <bitset>

class WAHBitVector;

extern
uint32_t bitcount(uint32_t v);

/***** Global bit operations on WAHBitVectors *****/
WAHBitVector operator& (const WAHBitVector& lhs, const WAHBitVector& rhs);
WAHBitVector operator| (const WAHBitVector& lhs, const WAHBitVector& rhs);

/********** Word by word iterator over WAHBitVector **********/
class WAHWordIterator {
  private:
    /* Iterator over the decoded words */
    uint32_t mWordIter, mWordIterRepeat, mWordIterCurr;
    const std::vector<uint32_t> & mBitWords;
    uint32_t mActiveWord, mActiveWordSize;

  public:
    WAHWordIterator(const std::vector<uint32_t> & bitWords,
                    uint32_t activeWord = 0, uint32_t activeWordSize = 0)
        : mBitWords(bitWords),
          mActiveWord(activeWord),
          mActiveWordSize(activeWordSize)
      { }

    void begin(uint32_t activeWord = 0, uint32_t activeWordSize = 0) {
        mWordIter = mWordIterRepeat = mWordIterCurr = 0;

        if(activeWordSize > 0) {
            mActiveWord = activeWord;
            mActiveWordSize = activeWordSize;
        }
    }

    /** returns next decoded 31-bit word */
    uint32_t getNextDecodedWord() {
        while(mWordIterRepeat) {
            mWordIterRepeat --;
            return mWordIterCurr; // return repeating decoded word
        }

        // try the active word if last
        if(mWordIter >= mBitWords.size()) {
            mActiveWordSize = 0; // hasMore() will return zero in next call
            return mActiveWord;
        }

        // decode next word
        if(mBitWords[mWordIter] > 0x7FFFFFFF) {
            // 0 or 1 fill
            mWordIterRepeat = (mBitWords[mWordIter] & 0x3FFFFFFF) - 1;
            mWordIterCurr = mBitWords[mWordIter] >= 0xC0000000 ? 0x7FFFFFFF : 0;
        }
        else {
            mWordIterRepeat = 0;
            mWordIterCurr = mBitWords[mWordIter];
        }
        mWordIter ++; // next undecoded word

        return mWordIterCurr;
    }

    bool hasMoreWords() {
        return (mWordIterRepeat || (mWordIter < mBitWords.size()) 
                || (mActiveWordSize > 0));
    }

};

/***** WAHBitVector class represents compressed bit vectors with WAH *****/
class WAHBitVector {
  private:
    uint32_t mSize;

    /* internal state variables */
    uint32_t              mNumWords;
    const std::bitset<32> ONES;
    std::bitset<32>       mActiveWord;
    uint32_t              mActiveWordIdx;
    uint32_t              mActiveWordSize;
    std::vector< uint32_t > mBitWords;

  public:
    WAHBitVector(uint32_t size) 
        : mSize(size),
          ONES(0x7FFFFFFF),
          mActiveWord(0),
          mActiveWordIdx(0),
          mIter(this->mBitWords)
      {
          mNumWords = size / 32 + 1;
          mBitWords.reserve(mNumWords);
      }

    /** number of bits */
    uint32_t size() const {
        return mSize;
    }

    /** number of 32-bit words required to store the bit vector */
    uint32_t getNumWords() const {
        return (this->mBitWords.size() + 1);
    }

    void appendUncompressed(uint32_t val) {
        this->mBitWords.push_back(val);
    }

    friend WAHBitVector operator& (const WAHBitVector& lhs, const WAHBitVector& rhs);
    friend WAHBitVector operator| (const WAHBitVector& lhs, const WAHBitVector& rhs);

    /**
     * \defgroup BitOperations Bit operation methods over the compressed bit vector.
     * @{ 
     */
    /**
     * In-place AND operator. Assumption vector lengths will be same. In-place
     * in the sense that the result will still be stored in this object although
     * internal representation is highly-likely to change, may even store the
     * result uncompressed.
     */
    WAHBitVector & operator&= (const WAHBitVector& rhs) {

        WAHBitVector result = (*this) & rhs;
        this->mActiveWord = result.mActiveWord;
        this->mActiveWordSize = result.mActiveWordSize;
        this->mBitWords.swap(result.mBitWords);
        return *this;

    }

    /* In-place OR operator. Assumption vector lengths will be same. */
    WAHBitVector & operator|= (const WAHBitVector& rhs) {

        WAHBitVector result = (*this) | rhs;
        this->mActiveWord = result.mActiveWord;
        this->mActiveWordSize = result.mActiveWordSize;
        this->mBitWords.swap(result.mBitWords);
        return *this;

    }

    /** Count set bits in a bit vector */
    uint32_t count() const {

        uint32_t count = 0;
        WAHWordIterator itr(this->mBitWords, 
                            this->mActiveWord.to_ulong(), 
                            this->mActiveWordSize);

        for(itr.begin(); itr.hasMoreWords(); ) {
            uint32_t word = itr.getNextDecodedWord();
            count += bitcount(word);
        }

        return count;

    }
    /**@}*/

    /** 
     * When building the bitvector, this method is called to inform about the
     * next set-bit's (bit with value 1) index.
     *
     * Subsequent calls of this method must have a greater idx than the previous
     * call. 
     * 
     * @param idx Index of the next set-bit in the bitvector.
     */
    void nextSetBitIndex(uint32_t idx) {
        uint32_t wordIdx = idx / 31;
        uint32_t bitIdx = idx % 31;

        if( wordIdx > mActiveWordIdx ) {
            if(mActiveWord == ONES) {
                // if the active word has all ones
                if(mBitWords.size() > 0
                   && (mBitWords.back() >= 0xC0000000) ) {
                    // if the last appended was compressed ones
                    mBitWords.back() ++;
                }
                else
                    mBitWords.push_back(0xC0000001);
            }
            else if(mActiveWord.none()) {
                // if the active word has all zeros
                if(mBitWords.size() > 0
                   && (mBitWords.back() >= 0x80000000 
                       && mBitWords.back() < 0xC0000000) ) {
                    // if the last appended was compressed zeros
                    mBitWords.back() ++;
                }
                else
                    mBitWords.push_back(0x80000001);
            }
            else {
                // append active word as a literal
                mBitWords.push_back(mActiveWord.to_ulong());
            }

            // append words with fill bit 0's
            if(mBitWords.back() >= 0x80000000 && mBitWords.back() < 0xC0000000){
                // last appended was all zeros
                mBitWords.back() += (wordIdx-mActiveWordIdx-1);
            }
            else if( (wordIdx-mActiveWordIdx) > 1 ) {
                // last appended was all ones or a literal 
                uint32_t zeros = 0x80000000 + (wordIdx-mActiveWordIdx-1);
                mBitWords.push_back(zeros);
            }

            mActiveWord.reset();
            mActiveWordSize = 0;
            mActiveWordIdx = wordIdx;
        }

        mActiveWord.set(bitIdx);
        mActiveWordSize = bitIdx+1;
    }

    /**
     * \defgroup BitIteration Iteration methods over the bits of the bit vector.
     * @{ 
     */

    /** Iterators over the bits */
    WAHWordIterator mIter;
    uint32_t mIterNextBit; //next iteration bit in curr word
    uint32_t mIterWord;

    void begin() {
        mIter.begin(this->mActiveWord.to_ulong(), this->mActiveWordSize);
        mIterNextBit = 31;
    }

    /** has more bits? */
    bool hasNext() {
        return (mIterNextBit < 31 || mIter.hasMoreWords());
    }

    /** return next bit, 0:false, 1:true */
    bool getNext() {

        if(mIterNextBit == 31) {
            mIterNextBit = 0;
            mIterWord = mIter.getNextDecodedWord();
        }

        ++ mIterNextBit;

        uint32_t ret = mIterWord & 0x1;
        mIterWord >>= 1;

        return ret;

    }
    /**@}*/
};

#endif /* _BITVECTOR_H_ */
