/**
 * Managing Large Data, Systems Group, ETH Zurich
 * Exercise 7
 * Bitmap Indexing, Query Processing and WAH Compression
 *
 * File        : queries.cpp
 * Author      : Cagri Balkesen, ETH Zurich, 2012-14-05
 *
 */

#include <string.h>
#include <stdio.h>
#include <iostream>

#include "queries.h"
#include "BitmapIndex.h"
#include "FieldType.h"

/******************************************************************************
 *
 * QUERY IMPLEMENTATIONS WITHOUT BITMAP INDEXES
 *
 *****************************************************************************/

/** 
 * Query 1: Execute the following TPCH query:
 *    SELECT count(*)
 *    FROM   lineitem
 *    WHERE  linenumber = 3 OR linenumber = 5;
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @return returns number of matching rows.
 */
int32_t
query1(uint32_t numrows, lineitem_t  *inputtable)
{
    /* Example implementation: */
    uint32_t count = 0;
    register uint32_t idx = 0;

    for(idx = 0; idx < numrows; ++idx){

        if(inputtable[idx].linenumber == 3 ||  inputtable[idx].linenumber == 5){

            count ++;

        }
    }

    return count;
}

/** 
 * Query 2: Execute the following TPCH query:
 *    SELECT count(*)
 *    FROM   lineitem
 *    WHERE  linenumber = 5      AND
 *           shipmode   = 'RAIL' AND
 *           quantity   = 10;
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @return returns number of matching rows.
 */
int32_t
query2(uint32_t numrows, lineitem_t  *inputtable)
{
    uint32_t count = 0;
    register uint32_t idx = 0;

    for(idx = 0; idx < numrows; ++idx){

        if( inputtable[idx].linenumber == 5
            && strcmp(inputtable[idx].shipmode, "RAIL") == 0
            && inputtable[idx].quantity == 10 ) {

            count ++;

        }

    }

    return count;
}

/** 
 * Query 3: Execute the following TPCH query:
 *    SELECT count(*)
 *    FROM   lineitem
 *    WHERE  linenumber IN (2, 4) AND
 *           shipmode   IN ('SHIP', 'TRUCK');
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @return returns number of matching rows.
 */
int32_t
query3(uint32_t numrows, lineitem_t  *inputtable)
{
    uint32_t count = 0;
    register uint32_t idx = 0;

    for(idx = 0; idx < numrows; ++idx){

        if( (inputtable[idx].linenumber == 2 
             || inputtable[idx].linenumber == 4 )
            && (strcmp(inputtable[idx].shipmode, "SHIP") == 0 
                || strcmp(inputtable[idx].shipmode, "TRUCK") == 0 ) ) {

            count ++;

        }
    }

    return count;
}

/** 
 * Query 4: Execute the following TPCH query:
 *    SELECT sum(extendedprice * (1.0-discount/100.0))
 *    FROM   lineitem
 *    WHERE  returnflag = 'R' AND
 *           shipmode IN ('REG AIR', 'AIR') AND
 *           quantity BETWEEN 5 AND 10;
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @return returns the sum value.
 */
double
query4(uint32_t numrows, lineitem_t  *inputtable)
{
    double sum = 0;
    register uint32_t idx = 0;

    for(idx = 0; idx < numrows; idx++){

        if( inputtable[idx].returnflag == 'R' 
            && (strcmp(inputtable[idx].shipmode, "REG AIR") == 0 
                || strcmp(inputtable[idx].shipmode, "AIR") == 0 )
            && (inputtable[idx].quantity >= 5 &&
                inputtable[idx].quantity <= 10) ) {

            sum += inputtable[idx].extendedprice *
                   (1.0-inputtable[idx].discount/100.0);
        }

    }

    return sum;
}

/******************************************************************************
 *
 * QUERY IMPLEMENTATIONS WITH BITMAP INDEXES
 *
 ******************************************************************************/

/** 
 * Query 1: Execute the following TPCH query using bitmap-index on linenumber
 *    SELECT count(*)
 *    FROM   lineitem
 *    WHERE  linenumber = 3 OR linenumber = 5;
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @param bitmap_idx 
 * 
 * @return returns number of matching rows.
 */
int32_t
query1_with_index(uint32_t numrows, lineitem_t  *inputtable, 
                  BitmapIndex<int32_t> & bitmap_idx_linenumber)
{
    // get all rows with linenumber = 3
    const BitVector & vec_ln3 = bitmap_idx_linenumber.getBitVector(3);
    // get all rows with linenumber = 5
    const BitVector & vec_ln5 = bitmap_idx_linenumber.getBitVector(5);
    // OR bitvectors
    BitVector res = vec_ln3 | vec_ln5;
    // return count
    return res.count();
}

/** 
 * Query 2: Execute the following TPCH query:
 *    SELECT count(*)
 *    FROM   lineitem
 *    WHERE  linenumber = 5      AND
 *           shipmode   = 'RAIL' AND
 *           quantity   = 10;
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @param bitmap_idx_linenumber bitmap index on linenumber
 * @param bitmap_idx_shipmode bitmap index on shipmode
 * @param bitmap_idx_quantity bitmap index on quantity
 * 
 * @return returns number of matching rows.
 */
int32_t
query2_with_index(uint32_t numrows, lineitem_t  *inputtable,
                  BitmapIndex<int32_t>  & bitmap_idx_linenumber,
                  BitmapIndex<char[11]> & bitmap_idx_shipmode,
                  BitmapIndex<int64_t>  & bitmap_idx_quantity)
{
    // get all rows with linenumber = 5
    const BitVector & vec_ln5 = bitmap_idx_linenumber.getBitVector(5);
    
    // get all rows with shipmode = 'RAIL'
    char rail[11];
    memset(rail, 0, 11);
    strcpy(rail, "RAIL");
    const BitVector & vec_smRAIL = bitmap_idx_shipmode.getBitVector(rail);

    // get all rows with quantity = 10
    const BitVector & vec_qt10 = bitmap_idx_quantity.getBitVector(10);

    // AND all vectors
    BitVector result = vec_ln5 & vec_smRAIL;
    result &= vec_qt10;

    // return count
    return result.count();
}

/** 
 * Query 3: Execute the following TPCH query:
 *    SELECT count(*)
 *    FROM   lineitem
 *    WHERE  linenumber IN (2, 4) AND
 *           shipmode   IN ('SHIP', 'TRUCK');
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @param bitmap_idx_linenumber bitmap index on linenumber
 * @param bitmap_idx_shipmode bitmap index on shipmode
 * 
 * @return returns number of matching rows.
 */
int32_t
query3_with_index(uint32_t numrows, lineitem_t  *inputtable,
                  BitmapIndex<int32_t>  & bitmap_idx_linenumber,
                  BitmapIndex<char[11]> & bitmap_idx_shipmode)
{
    // get all rows with linenumber = 2 or linenumber = 4
    const BitVector & vec_ln2 = bitmap_idx_linenumber.getBitVector(2);
    const BitVector & vec_ln4 = bitmap_idx_linenumber.getBitVector(4);

    // linenumber=2 OR linenumber=4
    BitVector vec_ln24 = vec_ln2 | vec_ln4;
    // get all rows with shipmode = 'SHIP' or 'TRUCK'
    char mode[11];
    memset(mode, 0, 11);
    strcpy(mode, "SHIP");
    const BitVector & vec_smSHIP = bitmap_idx_shipmode.getBitVector(mode);
    strcpy(mode, "TRUCK");
    const BitVector & vec_smTRUCK = bitmap_idx_shipmode.getBitVector(mode);

    // 'TRUCK' OR 'SHIP'
    BitVector vec_SHIPorTRUCK = vec_smSHIP | vec_smTRUCK;

    // AND them in-place
    vec_ln24 &= vec_SHIPorTRUCK;

    // return count
    return vec_ln24.count();
}

/** 
 * Query 4: Execute the following TPCH query:
 *    SELECT sum(extendedprice * (1.0-discount/100.0))
 *    FROM   lineitem
 *    WHERE  returnflag = 'R' AND
 *           shipmode IN ('REG AIR', 'AIR') AND
 *           quantity BETWEEN 5 AND 10;
 * 
 * @param numrows [in] number of rows in the input table.
 * @param inputtable [in] input table.
 * @param bitmap_idx_returnflag bitmap index on return flag
 * @param bitmap_idx_shipmode bitmap index on shipmode
 * @param bitmap_idx_quantity bitmap index on quantity
 * 
 * @return returns the sum value.
 */
double
query4_with_index(uint32_t numrows, lineitem_t  *inputtable,
                  BitmapIndex<char>     & bitmap_idx_returnflag,
                  BitmapIndex<char[11]> & bitmap_idx_shipmode,
                  BitmapIndex<int64_t>  & bitmap_idx_quantity)
{
    // get all rows with returnflag = 'R'
    const BitVector & vec_rfR = bitmap_idx_returnflag.getBitVector('R');

    // get all rows with shipmode = 'AIR' or 'REG AIR'
    char mode[11];
    memset(mode, 0, 11);
    strcpy(mode, "AIR");
    const BitVector & vec_smAIR = bitmap_idx_shipmode.getBitVector(mode);
    strcpy(mode, "REG AIR");
    const BitVector & vec_smREGAIR = bitmap_idx_shipmode.getBitVector(mode);
    // 'AIR' OR 'REG AIR'
    BitVector vec_AIRorREG = vec_smAIR | vec_smREGAIR;

    // get all rows with quantity >= 5 and quantity <= 10
    const BitVector & vec_qt10 = bitmap_idx_quantity.getBitVector(5);
    const BitVector & vec_qt11 = bitmap_idx_quantity.getBitVector(6);
    BitVector resultVec = vec_qt10 | vec_qt11;

    for(int i=7; i <= 10; i++)
        resultVec |= bitmap_idx_quantity.getBitVector(i);

    // AND with returnflag vector
    resultVec &= vec_rfR;

    // AND with shipmode vector
    resultVec &= vec_AIRorREG;

    double sum = 0;
    int i = 0;
    // Now iterate through final vector and fetch tuples with bit=1
    for(resultVec.begin(); resultVec.hasNext(); i++) {
        bool isSet = resultVec.getNext();

        if(isSet) {
            // fetch the row and aggregate
            sum += inputtable[i].extendedprice *
                   (1.0-inputtable[i].discount/100.0);
        }
    }

    return sum;
}
